@@ -63,14 +63,14 @@
 #endif
 #endif
 
-#define AE_LOCK(eventLoop)                                                                                             \
-    if ((eventLoop)->flags & AE_PROTECT_POLL) {                                                                        \
-        assert(pthread_mutex_lock(&(eventLoop)->poll_mutex) == 0);                                                     \
+#define AE_LOCK(eventLoop)                                         \
+    if ((eventLoop)->flags & AE_PROTECT_POLL) {                    \
+        assert(pthread_mutex_lock(&(eventLoop)->poll_mutex) == 0); \
     }
 
-#define AE_UNLOCK(eventLoop)                                                                                           \
-    if ((eventLoop)->flags & AE_PROTECT_POLL) {                                                                        \
-        assert(pthread_mutex_unlock(&(eventLoop)->poll_mutex) == 0);                                                   \
+#define AE_UNLOCK(eventLoop)                                         \
+    if ((eventLoop)->flags & AE_PROTECT_POLL) {                      \
+        assert(pthread_mutex_unlock(&(eventLoop)->poll_mutex) == 0); \
     }
 
 aeEventLoop *aeCreateEventLoop(int setsize) {
@@ -85,7 +85,7 @@ aeEventLoop *aeCreateEventLoop(int setsize) {
     if (eventLoop->events == NULL || eventLoop->fired == NULL) goto err;
     eventLoop->setsize = setsize;
     eventLoop->timeEventHead = NULL;
-    eventLoop->timeEventNextId = 0;
+    eventLoop->timeEventNextId = 1;
     eventLoop->stop = 0;
     eventLoop->maxfd = -1;
     eventLoop->beforesleep = NULL;
@@ -363,7 +363,7 @@ static int processTimeEvents(aeEventLoop *eventLoop) {
         }
 
         if (te->when <= now) {
-            int retval;
+            long long retval;
 
             id = te->id;
             te->refcount++;
