@@ -70,17 +70,24 @@ int anetGetError(int fd) {
     return sockerr;
 }
 
-int anetSetBlock(char *err, int fd, int non_block) {
+static int anetGetSocketFlags(char *err, int fd) {
     int flags;
 
-    /* Set the socket blocking (if non_block is zero) or non-blocking.
-     * Note that fcntl(2) for F_GETFL and F_SETFL can't be
-     * interrupted by a signal. */
     if ((flags = fcntl(fd, F_GETFL)) == -1) {
         anetSetError(err, "fcntl(F_GETFL): %s", strerror(errno));
         return ANET_ERR;
     }
 
+    return flags;
+}
+
+int anetSetBlock(char *err, int fd, int non_block) {
+    int flags = anetGetSocketFlags(err, fd);
+
+    if (flags == ANET_ERR) {
+        return ANET_ERR;
+    }
+
     /* Check if this flag has been set or unset, if so,
      * then there is no need to call fcntl to set/unset it again. */
     if (!!(flags & O_NONBLOCK) == !!non_block) return ANET_OK;
@@ -105,6 +112,21 @@ int anetBlock(char *err, int fd) {
     return anetSetBlock(err, fd, 0);
 }
 
+int anetIsBlock(char *err, int fd) {
+    int flags = anetGetSocketFlags(err, fd);
+
+    if (flags == ANET_ERR) {
+        return ANET_ERR;
+    }
+
+    /* Check if the O_NONBLOCK flag is set */
+    if (flags & O_NONBLOCK) {
+        return 0; /* Socket is non-blocking */
+    } else {
+        return 1; /* Socket is blocking */
+    }
+}
+
 /* Enable the FD_CLOEXEC on the given fd to avoid fd leaks.
  * This function should be invoked for fd's on specific places
  * where fork + execve system calls are called. */
@@ -474,6 +496,9 @@ static int anetTcpGenericConnect(char *err, const char *addr, int port, const ch
             continue;
         }
 
+        /* Enable TCP_NODELAY by default */
+        anetEnableTcpNoDelay(NULL, s);
+
         /* If we ended an iteration of the for loop without errors, we
          * have a connected socket. Let's return to the caller. */
         goto end;
@@ -614,37 +639,6 @@ int anetUnixServer(char *err, char *path, mode_t perm, int backlog, char *group)
     return s;
 }
 
-/* For some error cases indicates transient errors and accept can be retried
- * in order to serve other pending connections. This function should be called with the last errno,
- * right after anetTcpaccept or anetUnixAccept returned an error in order to retry them. */
-int anetRetryAcceptOnError(int err) {
-    /* This is a transient error which can happen, for example, when
-     * a client initiates a TCP handshake (SYN),
-     * the server receives and queues it in the pending connections queue (the SYN queue),
-     * but before accept() is called, the connection is aborted.
-     * in such cases we can continue accepting other connections. ß*/
-    if (err == ECONNABORTED)
-        return 1;
-
-#if defined(__linux__)
-    /* https://www.man7.org/linux/man-pages/man2/accept4.2 suggests that:
-     * Linux accept() (and accept4()) passes already-pending network
-       errors on the new socket as an error code from accept().  This
-       behavior differs from other BSD socket implementations.  For
-       reliable operation the application should detect the network
-       errors defined for the protocol after accept() and treat them like
-       EAGAIN by retrying.  In the case of TCP/IP, these are ENETDOWN,
-       EPROTO, ENOPROTOOPT, EHOSTDOWN, ENONET, EHOSTUNREACH, EOPNOTSUPP,
-       and ENETUNREACH. */
-    if (err == ENETDOWN || err == EPROTO || err == ENOPROTOOPT ||
-        err == EHOSTDOWN || err == ENONET || err == EHOSTUNREACH ||
-        err == EOPNOTSUPP || err == ENETUNREACH) {
-        return 1;
-    }
-#endif
-    return 0;
-}
-
 /* Accept a connection and also make sure the socket is non-blocking, and CLOEXEC.
  * returns the new socket FD, or -1 on error. */
 static int anetGenericAccept(char *err, int s, struct sockaddr *sa, socklen_t *len) {
@@ -693,6 +687,10 @@ int anetTcpAccept(char *err, int serversock, char *ip, size_t ip_len, int *port)
         if (ip) inet_ntop(AF_INET6, (void *)&(s->sin6_addr), ip, ip_len);
         if (port) *port = ntohs(s->sin6_port);
     }
+
+    /* Enable TCP_NODELAY by default */
+    anetEnableTcpNoDelay(NULL, fd);
+
     return fd;
 }
 
