@@ -36,7 +36,6 @@ typedef struct {
     size_t cow;
     monotime cow_updated;
     double progress;
-    size_t repl_output_bytes;
     childInfoType information_type; /* Type of information */
 } child_info_data;
 
@@ -65,7 +64,7 @@ void closeChildInfoPipe(void) {
 }
 
 /* Send save data to parent. */
-void sendChildInfoGeneric(childInfoType info_type, size_t keys, size_t repl_output_bytes, double progress, char *pname) {
+void sendChildInfoGeneric(childInfoType info_type, size_t keys, double progress, char *pname) {
     if (server.child_info_pipe[1] == -1) return;
 
     static monotime cow_updated = 0;
@@ -102,7 +101,6 @@ void sendChildInfoGeneric(childInfoType info_type, size_t keys, size_t repl_outp
 
     data.information_type = info_type;
     data.keys = keys;
-    data.repl_output_bytes = repl_output_bytes;
     data.cow = cow;
     data.cow_updated = cow_updated;
     data.progress = progress;
@@ -117,7 +115,7 @@ void sendChildInfoGeneric(childInfoType info_type, size_t keys, size_t repl_outp
 }
 
 /* Update Child info. */
-void updateChildInfo(childInfoType information_type, size_t cow, monotime cow_updated, size_t keys, size_t repl_output_bytes, double progress) {
+void updateChildInfo(childInfoType information_type, size_t cow, monotime cow_updated, size_t keys, double progress) {
     if (cow > server.stat_current_cow_peak) server.stat_current_cow_peak = cow;
 
     if (information_type == CHILD_INFO_TYPE_CURRENT_INFO) {
@@ -131,8 +129,6 @@ void updateChildInfo(childInfoType information_type, size_t cow, monotime cow_up
         server.stat_rdb_cow_bytes = server.stat_current_cow_peak;
     } else if (information_type == CHILD_INFO_TYPE_MODULE_COW_SIZE) {
         server.stat_module_cow_bytes = server.stat_current_cow_peak;
-    } else if (information_type == CHILD_INFO_TYPE_REPL_OUTPUT_BYTES) {
-        server.stat_net_repl_output_bytes += (long long)repl_output_bytes;
     }
 }
 
@@ -140,7 +136,7 @@ void updateChildInfo(childInfoType information_type, size_t cow, monotime cow_up
  * if complete data read into the buffer,
  * data is stored into *buffer, and returns 1.
  * otherwise, the partial data is left in the buffer, waiting for the next read, and returns 0. */
-int readChildInfo(childInfoType *information_type, size_t *cow, monotime *cow_updated, size_t *keys, size_t *repl_output_bytes, double *progress) {
+int readChildInfo(childInfoType *information_type, size_t *cow, monotime *cow_updated, size_t *keys, double *progress) {
     /* We are using here a static buffer in combination with the server.child_info_nread to handle short reads */
     static child_info_data buffer;
     ssize_t wlen = sizeof(buffer);
@@ -160,7 +156,6 @@ int readChildInfo(childInfoType *information_type, size_t *cow, monotime *cow_up
         *cow = buffer.cow;
         *cow_updated = buffer.cow_updated;
         *keys = buffer.keys;
-        *repl_output_bytes = buffer.repl_output_bytes;
         *progress = buffer.progress;
         return 1;
     } else {
@@ -175,12 +170,11 @@ void receiveChildInfo(void) {
     size_t cow;
     monotime cow_updated;
     size_t keys;
-    size_t repl_output_bytes;
     double progress;
     childInfoType information_type;
 
     /* Drain the pipe and update child info so that we get the final message. */
-    while (readChildInfo(&information_type, &cow, &cow_updated, &keys, &repl_output_bytes, &progress)) {
-        updateChildInfo(information_type, cow, cow_updated, keys, repl_output_bytes, progress);
+    while (readChildInfo(&information_type, &cow, &cow_updated, &keys, &progress)) {
+        updateChildInfo(information_type, cow, cow_updated, keys, progress);
     }
 }
