@@ -50,11 +50,11 @@ aofManifest *aofManifestCreate(void);
 void aofManifestFree(aofManifest *am);
 aofManifest *aofLoadManifestFromFile(sds am_filepath);
 
-#define ERROR(...)                                                                                                     \
-    {                                                                                                                  \
-        char __buf[1024];                                                                                              \
-        snprintf(__buf, sizeof(__buf), __VA_ARGS__);                                                                   \
-        snprintf(error, sizeof(error), "0x%16llx: %s", (long long)epos, __buf);                                        \
+#define ERROR(...)                                                              \
+    {                                                                           \
+        char __buf[1024];                                                       \
+        snprintf(__buf, sizeof(__buf), __VA_ARGS__);                            \
+        snprintf(error, sizeof(error), "0x%16llx: %s", (long long)epos, __buf); \
     }
 
 static char error[1044];
@@ -556,12 +556,6 @@ int redis_check_aof_main(int argc, char **argv) {
         goto invalid_args;
     }
 
-    /* Check if filepath is longer than PATH_MAX */
-    if (strnlen(filepath, PATH_MAX + 1) > PATH_MAX) {
-        printf("Error: filepath is too long (exceeds PATH_MAX)\n");
-        goto invalid_args;
-    }
-
     /* In the glibc implementation dirname may modify their argument. */
     memcpy(temp_filepath, filepath, strlen(filepath) + 1);
     dirpath = dirname(temp_filepath);
