@@ -1,18 +1,18 @@
-# Cluster test suite. Copyright (C) 2014 Salvatore Sanfilippo antirez@gmail.com
+# Cluster test suite. Copyright (C) 2014 Redis Ltd.
 # This software is released under the BSD License. See the COPYING file for
 # more information.
 
 cd tests/cluster
 source cluster.tcl
 source ../instances.tcl
-source ../../support/cluster.tcl ; # Redis Cluster client.
+source ../../support/cluster.tcl ; # Cluster client.
 
 set ::instances_count 20 ; # How many instances we use at max.
 set ::tlsdir "../../tls"
 
 proc main {} {
     parse_options
-    spawn_instance redis $::redis_base_port $::instances_count {
+    spawn_instance valkey $::valkey_base_port $::instances_count {
         "cluster-enabled yes"
         "appendonly yes"
         "enable-protected-configs yes"
