@@ -1,4 +1,4 @@
-# Sentinel test suite. Copyright (C) 2014 Salvatore Sanfilippo antirez@gmail.com
+# Sentinel test suite. Copyright (C) 2014 Redis Ltd.
 # This software is released under the BSD License. See the COPYING file for
 # more information.
 
@@ -19,7 +19,7 @@ proc main {} {
         "enable-debug-command yes"
     } "../tests/includes/sentinel.conf"
 
-    spawn_instance redis $::redis_base_port $::instances_count {
+    spawn_instance valkey $::valkey_base_port $::instances_count {
         "enable-protected-configs yes"
         "enable-debug-command yes"
         "save ''"
