@@ -142,6 +142,31 @@ start_server {tags {"expire"}} {
         list $e $f
     } {somevalue {}}
 
+    test {EXPIRE / EXPIREAT / PEXPIRE / PEXPIREAT Expiration time is already expired} {
+        r flushall
+        r config resetstat
+
+        r set x somevalue
+        r expire x -1
+        assert_equal {0} [r exists x]
+        assert_equal {1} [s expired_keys]
+
+        r set x somevalue
+        r expireat x [expr [clock seconds] - 1]
+        assert_equal {0} [r exists x]
+        assert_equal {2}  [s expired_keys]
+
+        r set x somevalue
+        r pexpire x -1000
+        assert_equal {0} [r exists x]
+        assert_equal {3} [s expired_keys]
+
+        r set x somevalue
+        r pexpireat x [expr [clock milliseconds] - 1000]
+        assert_equal {0} [r exists x]
+        assert_equal {4} [s expired_keys]
+    }
+
     test {TTL returns time to live in seconds} {
         r del x
         r setex x 10 somevalue
@@ -832,6 +857,98 @@ start_server {tags {"expire"}} {
         close_replication_stream $repl
         assert_equal [r debug set-active-expire 1] {OK}
     } {} {needs:debug}
+
+    test {import-source can be closed when import-mode is off} {
+        r config set import-mode no
+        assert_error "ERR Server is not in import mode" {r client import-source on}
+
+        r config set import-mode yes
+        assert_equal [r client import-source on] {OK}
+        assert_match {*flags=I*} [r client list id [r client id]]
+
+        r config set import-mode no
+        assert_equal [r client import-source off] {OK}
+        assert_match {*flags=N*} [r client list id [r client id]]
+    }
+
+    test {Import mode should forbid active expiration} {
+        r flushall
+
+        r config set import-mode yes
+        assert_equal [r client import-source on] {OK}
+
+        r set foo1 bar PX 1
+        r set foo2 bar PX 1
+        after 10
+
+        assert_equal [r dbsize] {2}
+
+        assert_equal [r client import-source off] {OK}
+        r config set import-mode no
+
+        # Verify all keys have expired
+        wait_for_condition 40 100 {
+            [r dbsize] eq 0
+        } else {
+            fail "Keys did not actively expire."
+        }
+    }
+
+    test {Import mode should forbid lazy expiration} {
+        r flushall
+        r debug set-active-expire 0 
+
+        r config set import-mode yes
+        assert_equal [r client import-source on] {OK}
+
+        r set foo1 1 PX 1
+        after 10
+
+        r get foo1
+        assert_equal [r dbsize] {1}
+
+        assert_equal [r client import-source off] {OK}
+        r config set import-mode no
+
+        r get foo1
+
+        assert_equal [r dbsize] {0}
+
+        assert_equal [r debug set-active-expire 1] {OK}
+    } {} {needs:debug}
+
+    test {Client can visit expired key in import-source state} {
+        r flushall
+
+        r config set import-mode yes
+
+        r set foo1 1 PX 1
+        after 10
+
+        # Normal clients cannot visit expired key.
+        assert_equal [r get foo1] {}
+        assert_equal [r ttl foo1] {-2}
+        assert_equal [r dbsize] 1
+
+        # Client can visit expired key when in import-source state.
+        assert_equal [r client import-source on] {OK}
+        assert_equal [r ttl foo1] {0}
+        assert_equal [r get foo1] {1}
+        assert_equal [r incr foo1] {2}
+        assert_equal [r randomkey] {foo1}
+        assert_equal [r scan 0 match * count 10000] {0 foo1}
+        assert_equal [r keys *] {foo1}
+
+        assert_equal [r client import-source off] {OK}
+        r config set import-mode no
+
+        # Verify all keys have expired
+        wait_for_condition 40 100 {
+            [r dbsize] eq 0
+        } else {
+            fail "Keys did not actively expire."
+        }
+    }
 }
 
 start_cluster 1 0 {tags {"expire external:skip cluster"}} {
@@ -847,7 +964,7 @@ start_cluster 1 0 {tags {"expire external:skip cluster"}} {
 
         # hashslot(foo) is 12182
         # fill data across different slots with expiration
-        for {set j 1} {$j <= 100} {incr j} {
+        for {set j 1} {$j <= 1000} {incr j} {
             r psetex "{foo}$j" 500 a
         }
         # hashslot(key) is 12539
@@ -858,7 +975,7 @@ start_cluster 1 0 {tags {"expire external:skip cluster"}} {
         r debug dict-resizing 0
 
         # delete data to have lot's (99%) of empty buckets (slot 12182 should be skipped)
-        for {set j 1} {$j <= 99} {incr j} {
+        for {set j 1} {$j <= 999} {incr j} {
             r del "{foo}$j"
         }
 
@@ -884,7 +1001,9 @@ start_cluster 1 0 {tags {"expire external:skip cluster"}} {
         r debug dict-resizing 1
 
         # put some data into slot 12182 and trigger the resize
+        # by deleting it to trigger shrink
         r psetex "{foo}0" 500 a
+        r del "{foo}0"
 
         # Verify all keys have expired
         wait_for_condition 400 100 {
